#-----------------------------------------------------------------------------
#
# Thomas Thomassen
# thomas[at]thomthom[dot]net
#
#-----------------------------------------------------------------------------

# Wrapper class for Model.raytest with enhanced features.
#
# @since 2.0.0
class TT::Ray
  
  # @overload new(ray)
  #   @param [Array<Sketchup::Point3d,Sketchup::Vector3d>] value
  # @overload new(ray)
  #   @param [Array<Sketchup::Point3d,Sketchup::Point3d>] value
  # @overload new(point, vector)
  #   @param [Sketchup::Point3d] point
  #   @param [Sketchup::Vector3d] vector
  # @overload new(point1, point2)
  #   @param [Sketchup::Point3d] point1
  #   @param [Sketchup::Point3d] point2
  #
  # @since 2.0.0
  def initialize( *args )
    if args.size == 1
      @ray = args[0]
    elsif args.size == 2
      raise ArgumentError unless args[0].is_a?( Geom::Point3d )
      if args[1].is_a?( Geom::Point3d )
        @ray = [ args[0], args[0].vector_to(args[1]) ]
      elsif args[1].is_a?( Geom::Vector3d )
        @ray = [ args[0], args[1] ]
      else
        raise ArgumentError
      end
    else
      raise ArgumentError
    end
  end
  
  
  # Gets the origin of the ray.
  #
  # @return [Sketchup::Point3d]
  # @since 2.0.0
  def origin
    @ray[0]
  end
  
  # Sets the origin of the ray.
  #
  # @param [Sketchup::Point3d] point
  #
  # @return [Sketchup::Point3d]
  # @since 2.0.0
  def origin=(point)
    @ray[0] = point
  end
  
  
  # Gets the direction of the ray.
  #
  # @return [Sketchup::Vector3d]
  # @since 2.0.0
  def direction
    @ray[1]
  end
  
  # Sets the origin of the ray.
  #
  # @overload direction=(point)
  #   @param [Sketchup::Point3d] point
  # @overload direction=(vector)
  #   @param [Sketchup::Vector3d] vector
  #
  # @return [Sketchup::Vector3d]
  # @since 2.0.0
  def direction=(value)
    if value.is_a?( Geom::Point3d )
      @ray[1] = @ray[0].vector_to(value)
    elsif value.is_a?( Geom::Vector3d )
      @ray[1] = value
    else
      raise ArgumentError
    end
  end
  
  
  # @return [Array<Sketchup::Point3d, Sketchup::Vector3d>]
  # @since 2.0.0
  def to_a
    return @ray
  end
  
  # @return [String]
  # @since 2.0.0
  def to_s
    return "#{@ray[0]}, #{@ray[1]}"
  end
  
  
  # @return [String]
  # @since 2.0.0
  def inspect
    return "Ray(#{@ray[0].inspect}, #{@ray[1].inspect})"
  end
  
  
  # Test the ray and do not stop on hidden entities. Option to stop at ground
  # level.
  #
  # Version 2.1.1 Works around a bug with model.raytest in SU8.0.
	#
	# @param [Sketchup::Model] model
	# @param [Boolean] stop_at_ground
	#
	# @return [Array<Point3d, Array>, nil]
	# @since 2.0.0 - fixed in 2.1.0
  def test( model = Sketchup.active_model, stop_at_ground = false )
    ground = [ORIGIN, Z_AXIS]
    ray = @ray
    
    while true
    
      # Shot ray
      test = model.raytest(ray)
      
      # Check if the ray should stop at ground
      if test.nil?
        # Stop the ray at the ground if it misses any geometry.
        if stop_at_ground
          # Shoot ray
          pt = Geom.intersect_line_plane( @ray, ground )
          return nil if pt.nil?
          v = @ray[0].vector_to(pt)
          if v.valid? && v.samedirection?(@ray[1])
            return [ pt, [] ]
          else
            return nil
          end
        else
          return nil
        end
      end
      
      # Verify visibility
      point, path = test
      
      # (!) SketchUp 8.0 (Initial Release) introduced a bug where rays passing 
      # through the XY, XZ or YZ plane would return 0,0,0 and a CPoint as path
      # when there is a CPoint in model.active_entities.
      # http://forums.sketchucation.com/viewtopic.php?f=180&t=30558
      #
      # Rays should never return a ConstructionPoint - use this as a method to
      # detect when the bug occurs.
      valid_ray = !path.first.is_a?( Sketchup::ConstructionPoint )
      
      if valid_ray && path.all? { |e| e.visible? && e.layer.visible? }
        return test
      else
        # raytest stopped on a hidden entity - create a new ray from that point
        # and let the loop continue until it hits something visible, or nothing.
        if valid_ray
          ray = [ point, @ray[1] ]
        else
          # Workaround for SU8.0's raytest bug.
          # (!) Intersect with XY,XZ and YZ planes to determine where the ray
          # should restart from.
          new_point = nil
          [X_AXIS,Y_AXIS,Z_AXIS].each { |v|
            plane = [ ORIGIN, v ]
            pt = Geom.intersect_line_plane( ray, plane )
            next if pt.nil?
            direction = ray.first.vector_to( pt )
            new_point = pt if direction.samedirection?( ray.last )
          }
          if new_point.nil?
            return nil
          else
            ray = [ new_point, @ray[1] ]
          end
        end # if valid_ray
      end
      
    end # while
  end
  
end # class TT::Ray